/**
 * @file Power.h
 * @author Phillip Kühne
 * @brief This component provides utilities for keeping track of power usage
 * consumption.
 * @version 0.1
 * @date 2024-11-23
 */

#include "PowerScheduler.h"
#ifndef Power_h
#define Power_h

enum TaskResumptionReason { POWER_AVAILABLE, TIMEOUT };

class Power {

public:
  static void begin(void);
  Power();
  uint16_t getFreeCurrentBudget(void);
  /// @brief Request an allowance of a certain number of milliwatts from the
  /// power scheduler
  /// @param neededPower the amount of power we want to be accounted for (in mW)
  /// @return whether the power could be successfully allocated
  static bool
  tryAccquireCurrentAllowance(PowerParameters::PowerConsumers consumer,
                              uint16_t neededCurrent,
                              uint16_t requestedDurationMs);
  /// @brief "Return" the power consumed by an active consumer
  /// @param neededPower the amount of power to return (in mW)
  /// @return whether the power
  static void releaseCurrent(PowerParameters::PowerConsumers consumer);

  /// @brief Wait for a certain amount of power to be available
  /// @param neededPower the amount of power we want to be accounted for (in mW)
  /// @param TicksToWait the amount of time to wait for the power to become
  /// available
  /// @return whether the power could be successfully allocatedy
  static bool waitForCurrentAllowance(PowerParameters::PowerConsumers consumer,
                                      uint16_t neededCurrent,
                                      TickType_t TicksToWait,
                                      uint16_t requestedDurationMs);
  /// @brief Put the ESP32 into deep sleep mode, without a method to wake up
  /// again. Basically this is a shutdown.
  static void beginPermanentDeepSleep(void);

  /// @brief Get battery voltage measurement.
  /// @return Battery Terminal Voltage in Volts
  static float getBatteryVoltage();

  /// @brief Get estimated battery charge state as percentage
  /// @return Battery charge state in percent
  static int getBatteryChargePercent();

  /// @brief Get estimated battery charge state as percentage based on
  // voltage directly
  /// @return Battery charge state in percent
  static int getBatteryVoltageChargePercent();

  /// @brief Get estimated battery charge state as coulombs
  /// @return Battery charge state in coulombs
  static float getBatteryChargeCoulombs();

  /// @brief get available current (after voltage conversion and efficiency
  /// losses, referencing 1C discharge)
  /// @return available current in milliamps
  static float getMax3V3Current();

protected:
  /// @brief PowerScheduler instance to manage power consumption
  static PowerScheduler *powerScheduler;

  /// @brief update Power State
  static void updatePowerStateHandler();

  /*
   * Power State
   */

  /// @brief last time of power state update
  static TickType_t lastPowerStateUpdate;

  /// @brief remaining Charge in coulombs
  static float coloumbsRemaining;

  /// @brief remaining Charge in percent
  static int percentRemaining;

  /// @brief Circular array of last calculated values for current state of
  /// charge
  static float lastSOC[PowerParameters::Battery::AVERAGING_SAMPLES];
  static int latestSoCIndex;

  /// @brief Add calculated value to circular array, pushing out oldest value
  static void addSoCSample(float soc);
};

extern Power power;

#endif // Power